# !/usr/bin/env python

# This file was originally written by jirihnidek (username on GitHub), here are some links
# regarding the original code:
#
# Author's Website:        https://jirihnidek.github.io/
# Author Gist Page:        https://gist.github.com/jirihnidek
# Source of Original Code: https://gist.github.com/jirihnidek/3f5d36636198e852280f619847d22d9e

"""Aliases for argparse positional arguments."""

import argparse


class SubparserActionAliases(argparse._SubParsersAction):
    class _AliasedPseudoAction(argparse.Action):

        def __init__(self, name, aliases, help):
            dest = name
            if aliases:
                dest += f" ({','.join(aliases)})"
            sup = super(SubparserActionAliases._AliasedPseudoAction, self)
            sup.__init__(option_strings=[], dest=dest, help=help)

    def add_parser(self, name, **kwargs):
        """
        The add_parser function adds a parser to the subparsers object.
        It adds it to the _name_parser_map dictionary, which maps command \
        name aliases to parsers. It also adds an entry in _choices_actions, which is used by argparse
        to create help text for each of these commands.

        Args:
            name:
                Specify the name of the command
            **kwargs:
                Pass a dictionary of keyword arguments to the add_parser function

        Returns:
            The parser object, which we can then use to add arguments to
        """
        if 'aliases' in kwargs:
            aliases = kwargs['aliases']
            del kwargs['aliases']
        else:
            aliases = []

        parser = super(SubparserActionAliases, self).add_parser(name, **kwargs)

        # Make the aliases work.
        for alias in aliases:
            self._name_parser_map[alias] = parser
        # Make the help text reflect them, first removing old help entry.
        if 'help' in kwargs:
            help = kwargs.pop('help')
            self._choices_actions.pop()
            pseudo_action = self._AliasedPseudoAction(name, aliases, help)
            self._choices_actions.append(pseudo_action)

        return parser



from inspyre_toolbox.version import parse_version
from inspyre_toolbox.common.meta import *


__PROG__ = 'InspyreToolbox'
__VERSION__: str = parse_version()
__AUTHORS__: list[tuple[str, str]] = AUTHORS



#  Copyright (c) 2021. Taylor-Jayde Blackstone <t.blackstone@inspyre.tech> https://inspyre.tech
import contextlib
import ctypes
import os
from datetime import datetime

import psutil
from pypattyrn.behavioral.null import Null

import inspyre_toolbox.settings as it_settings
from inspyre_toolbox.core_helpers.logging import add_isl_child, force_lowkey_log_name, ISL as InspyLogger
from inspyre_toolbox.humanize import Numerical
from inspyre_toolbox.proc_man.errors import NoFoundProcessesError


fts = datetime.fromtimestamp

ISL = InspyLogger('InspyreToolBox.ProcMan', it_settings.log_level.upper())

LOG = ISL


class Colors(object):
    def __init__(self, return_null=True):
        """
        Initializes a Colors object.

        Args:
            return_null (bool, optional):
                If True, instantiate the Colors object with Null values for each color attribute.

                If False, instantiate the Colors object with actual color values.Defaults to True.
        """
        if return_null:
            self.red       = Null()
            self.yellow    = Null()
            self.blue      = Null()
            self.green     = Null()
            self.end_color = Null()
        else:
            from inspyred_print import Color, Format
            color = Color()
            fmt = Format()
            self.red       = color.red
            self.yellow    = color.yellow
            self.blue      = color.blue
            self.green     = color.green
            self.end_color = fmt.end_mod


def is_admin() -> bool:
    """
    Checks if the current user has administrative privileges.

    Returns:
        bool:
            True if the current user has administrative privileges, False otherwise.
    """
    try:
        _is_admin = (os.getuid() == 0)
    except AttributeError:
        _is_admin = ctypes.windll.shell32.IsUserAnAdmin() != 0
    return _is_admin


def find_all_by_name(name, case_sensitive=False, inspy_logger_device=None, on_the_dl=False, colorful_logging=False):
    """

    Find a running program by name.

    Args:
        name (str):
            The name of the program you're looking for.

        case_sensitive (bool):
            Should the function care about letter casing when searching for the given program? (
            Optional) - Defaults to bool(False)

        inspy_logger_device (inspy_logger.InspyLogger().device):
            The instantiated device object from the inspy_logger package. (Optional, defaults to None)

            Note:
                If a value other than NoneType is passed to this parameter, the function will check the
                InspyLogger.device attribute 'started' to see if the logger has already been started elsewhere. One
                of three situations could result from this:

                    1) The attribute 'started' exists, and evaluates to bool(False):

                       In this case the function will attempt to start the log-device by calling it's 'start' method
                       before calling 'add_child' on the log-device in the hopes of receiving it's own logging object.

                    2) The attribute 'started' exists and evaluates to bool(False):

                       In this case the function will attempt to call 'add_child' on the log-device in the hopes of
                       receiving it's own logging object.

                    3) The attribute 'started' does not exist:

                       In this case, the function will receive a raised AttributeError from the passed object. If
                       this occurs; the function will raise a more relevant 'InvalidLogDeviceError'

        on_the_dl (bool):
            Should InspyreToolbox leave itself out of the the logging name? (Optional, Defaults to bool(False))

        colorful_logging (bool):
            Should the function log colorfully? (Optional, Defaults to bool(False))


    Returns:
        procs_found (list): A list of the currently running processes that match the given name.

    """

    isl_dev = inspy_logger_device

    prefix = '' if on_the_dl else 'InspyreToolbox.'
    log_name = f'{prefix}proc_man.find_by_name'

    log = add_isl_child(log_name, isl_dev)

    r_null = force_lowkey_log_name or not colorful_logging
    colors = Colors(return_null=not colorful_logging)

    procs_found = []

    # If we're not being case-sensitive we'll lower the case on 'name'
    if not case_sensitive:
        log.debug('Searching in case insensitive mode!')
        name = name.lower()

    log.debug(f'Search query: {name}')

    # Iterate through the process list
    for proc in psutil.process_iter():
        with contextlib.suppress(psutil.NoSuchProcess, psutil.AccessDenied, psutil.ZombieProcess):
            # Gather the information of all the currently running processes
            proc_info = proc.as_dict(
                    attrs=['pid', 'name', 'create_time', 'username'])
            log.debug(proc_info)

            # If we're not caring about case let's go ahead and make both
            # things we're comparing lowercase.
            proc_name = proc_info['name'] if case_sensitive else proc_info['name'].lower()
            if name in proc_name:
                log.debug(f'{colors.green}Found match: {colors.blue}{proc_info["name"]}'
                          f'({colors.yellow}{proc_info["pid"]}{colors.blue})')

                # Add find to the list we're returning to the caller.
                procs_found.append(proc_info)

                log.debug(
                        f'{colors.green}Added {proc_info["pid"]} to found process list')
    num_found = Numerical(len(procs_found), noun='process')

    log.debug(f'Found: {colors.yellow}{num_found.count_noun()}')

    # Return found process list to caller
    return procs_found


def kill_all_in_list(kill_list, inspy_logger_device=None, on_the_dl=False, colorful_logging=False):
    """

    When provided a list of processes, kill each process in the list.

    Args:
        kill_list:
            A list of processes in the format of a dictionary

        inspy_logger_device:
            An instantiated inspy-logger device.

        on_the_dl:
            Don't include 'InspyreToolbox' in the name of the logger.

        colorful_logging:
            Should logging be colorful?

    Returns:
        None

    """

    prefix = '' if on_the_dl else 'InspyreToolbox.'
    log_name = f'{prefix}ProcMan.kill_all_in_list'

    log = add_isl_child(log_name, ISL.device)

    log.debug('Logging active')

    colors = Colors(return_null=not colorful_logging)

    missing = 0

    for proc in kill_list:
        log.debug(f'Killing {colors.yellow}{proc["pid"]}')
        os.kill(proc['pid'], 2)
        # try:
        #         subprocess.check_output("Taskkill /PID %d /F" % proc['pid'])
        #     if os.name == 'nt':
        #     else:
        #         os.kill(proc['pid'], 2)
        # except subprocess.CalledProcessError:
        #     missing += 1
        #     if missing >= 20:
        #         kill_all_by_name(proc['name'])
        #         break
        # log.debug(
        #     f'{colors.green}Kill signal sent to {colors.yellow}{proc["pid"]}')


def kill_all_by_name(
        name,
        case_sensitive=False,
        inspy_logger_device=None,
        on_the_dl=False,
        colorful_logging=False
):
    """

    Kill all currently running programs (and all instances thereof) which have a name containing the string given as a
    value to the 'name' parameter.

    Note:
        Unless you are running this as administrator, any programs that you don't have permission to kill will not be
        killed.

        IF YOU ARE RUNNING AS ADMINISTRATOR EVERY PROCESS THAT MATCHES WILL BE KILLED REGARDLESS OF WHO INSTANTIATED
        THE PROCESS!!!!

    Args:
        name (str):
            The name of the program you'd like to kill all instances of. Whatever
            you provide here will be used as a query to find all programs that have
            a name that contains this string.

        case_sensitive (bool):
            Tell the function that the letter casing is irrelevant to finding matching
             programs. (Optional, defaults to False)

        inspy_logger_device (inspy_logger.InspyLogger().device):
            The instantiated device object from the inspy_logger package. (Optional, defaults to None)

            Note:
                If a value other than NoneType is passed to this parameter, the function will check the
                InspyLogger.device attribute 'started' to see if the logger has already been started elsewhere. One
                of three situations could result from this:

                    1) The attribute 'started' exists, and evaluates to bool(False):

                       In this case the function will attempt to start the log-device by calling it's 'start' method
                       before calling 'add_child' on the log-device in the hopes of receiving it's own logging object.

                    2) The attribute 'started' exists and evaluates to bool(False):

                       In this case the function will attempt to call 'add_child' on the log-device in the hopes of
                       receiving it's own logging object.

                    3) The attribute 'started' does not exist:

                       In this case, the function will receive a raised AttributeError from the passed object. If
                       this occurs; the function will raise a more relevant 'InvalidLogDeviceError'

        on_the_dl (bool):
            Should InspyreToolbox leave itself out of the the logging name? (Optional, Defaults to bool(False))

        colorful_logging (bool):
            Should the function log colorfully? (Optional, Defaults to bool(False))

    Returns:
        None

    """
    from time import sleep

    # Get a logger for logging in logging is enabled.
    log = add_isl_child('proc_man.kill_all_by_name', inspy_logger_device)

    # Get colors for colorful logging if colorful_logging is bool(True)
    color = Colors(return_null=not colorful_logging)

    log.debug(f'{color.yellow}Checking to see if there are, any running programs with "'
              f'{color.green}{name}{color.yellow}" in their name.')

    # Get a list of all the currently running processes that contains 'name' string.
    procs = find_all_by_name(name, case_sensitive,
                             inspy_logger_device, on_the_dl, colorful_logging)

    log.debug(procs)

    sleep(5)

    # Get the number of running processes found.
    num_running = Numerical(len(procs), noun='process')

    log.debug(f'Found {num_running.count_noun()}')

    # Gather the username of the current user.
    username = os.getlogin()

    # Find out if they are an administrator or not.
    __is_admin = is_admin()

    started_by_user = []

    if not __is_admin:
        for proc in procs:
            log.debug(proc['username'])
            if proc['username'].endswith(username):
                started_by_user.append(proc)

    started_by_user.reverse()

    kill_all_in_list(started_by_user)



# ==============================================================================
#  Copyright (c) Inspyre Softworks 2022.                                       =
#                                                                              =
#  Author:                 T. Blackstone                                       =
#  Author Email:    <t.blackstone@inspyre.tech>                                =
#  Created:              2/10/22, 9:49 PM                                      =
# ==============================================================================

from pypattyrn.behavioral.null import Null

force_lowkey_log_name = False
"""

This variable should be a boolean indicating whether or not InspyreToolbox should leave itself out when it's
functions name their child loggers.

Note:


For Example:

    Let's look at how 'inspyre_toolbox.proc_man.find_all_by_name' would name it's loggers in each of the situations we
    could find ourselves in;
   
    Note:
        [ROOT_LOGGER_NAME] is just a placeholder for whatever the name of the root logging device is.

    Case 1 (lowkey_log_name evaluates to bool(False)):
        
        [ROOT_LOGGER_NAME].InspyreToolbox.proc_man.find_all_by_name
        
    Case 2 (lowkey_log_name evaluates to any value seen as bool(True)):
        
        [ROOT_LOGGER_NAME].proc_man.find_all_by_name
        

"""
INSPY_LOG_LEVEL = 'info'
from inspy_logger import Logger as ISL
import inspyre_toolbox.settings as it_settings

prog = 'InspyreToolbox'

root_ISL = ISL(prog, it_settings.log_level.upper())

ROOT_ISL_DEVICE = root_ISL

PROG_LOGGERS = {}


class InvalidLogDeviceError(Exception):

    def __init__(self):
        """

        Raised when one of the functions in InspyreToolbox are given a value to "inspy_logger_device" and it's found
        that the object provided is not properly configured, thus receiving an AttributeError when trying to check
        inspy_logger_device.started for a bool() value.

        Note:
            'inspy_logger_device.started' will exist in all cases where the logger device was given a name and
            log-level on instantiation (or later through setter methods). If the logging device hasn't been set up,
            an AttributeError will cause the following functions to raise InvalidLogDeviceError:

                * find_by_name
                * kill_all_by_name

        """
        self.message = 'The passed device does not appear to have been initialized, or is not an InspyLogger device!'


class Manifest(list):

    def __init__(self, *arg, **kw):
        # assert isinstance(comp, dict), '"comp" requires a dictionary!'
        #
        # for x, y in comp.items():
        #     assert isinstance(x, str), f"Invalid key {x}. Please input a string, not {type(x).__name__}"
        #     assert type(y).__name__ == 'Logger', f"Invalid value, the value should be a 'Logger' device. Instead got " \
        #                                          f"type {type(y).__name__}"
        #
        # self.__comp = comp
        # self.__get_comp_map()
        super(Manifest, self).__init__(*arg, **kw)
        self.__slots__ = ()

    def __get_comp_map(self):
        self.__comp_map = [(x, y1) for x, y in self.__comp.items() for y1 in y]

    @staticmethod
    def __sum_lists(lists):
        lt = []
        for lst in lists:
            lt += lst
        return lt

    def __get_all(self):
        return self.__sum_lists(map(list, self.__comp.values()))

    def __repr__(self):
        return str(self.__get_all())


def add_isl_child(name, isl_device=ROOT_ISL_DEVICE):
    """

    Manage using an InspyLogger device with Inspyre Toolbox.

    Args:
        name:
            The name you'd like for your log-device.

        isl_device (inspy_logger.InspyLogger.device | None):
            An instantiated inspy-logger device.

    Returns:
        The log device.

    """
    return ROOT_ISL_DEVICE.get_child(name)



import subprocess
import sys


def up_to_date(name):
    latest_version = str(
        subprocess.run(
            [sys.executable, '-m', 'pip', 'install', f'{name}==random'],
            capture_output=True,
            text=True,
        )
    )

    latest_version = latest_version[latest_version.find('(from versions:') + 15:]
    latest_version = latest_version[:latest_version.find(')')]
    latest_version = latest_version.replace(' ', '').split(',')[-1]

    current_version = str(
        subprocess.run(
            [sys.executable, '-m', 'pip', 'show', f'{name}'],
            capture_output=True,
            text=True,
        )
    )

    current_version = current_version[current_version.find('Version:') + 8:]
    current_version = current_version[:current_version.find(
            '\\n')].replace(' ', '')

    return latest_version == current_version



"""

This file contains custom exceptions for Inspyre_Toolbox.live_timer.

"""
from inspyre_toolbox.core_helpers.debugging import who_rang


class TimerNotRunningError(Exception):
    default_message = 'This timer is not running!'

    def __init__(self, message=default_message, skip_print=False, caller=who_rang()):
        self.actual_caller = who_rang()
        self.caller_msg = f'Erroneous call from: {self.actual_caller}'

        if self.actual_caller != caller:
            self.caller_msg += f' which claims it was called by {caller}'

        msg = f'{self.default_message} {self.caller_msg}'

        if message != self.default_message:
            msg += f"Further information from caller: {message}"

        self.message = msg

        if not skip_print:
            print(self.message)


class TimerNotStartedError(TimerNotRunningError):
    default_message = 'Timer has not yet been started!'

    def __init__(self, message=default_message, skip_print=False, caller=who_rang(), **kwargs):
        super(TimerNotStartedError, self).__init__(kwargs)



"""
File:
    inspyre_toolbox/cli/ist_bytes_converter/helpers

Author:
    Inspyre-Softworks
    Taylor-Jayde <t.blackstone@inspyre.tech>

Description:
    Helpers for the `ist-bytes-converter` CLI-App.

Functions:
    clean_unit_str:
        Cleans and standardizes the  unit string by removing plural 's' and converting to lowercase.

Variables:
    __all__:

"""


__all__ = [
    'clean_unit_str',
    'convert_if_whole_number'
]


def clean_unit_str(unit_str):
    """
    Cleans and standardizes the unit string by removing plural 's' and converting to lowercase.

    Parameters:
        unit_str (str):
            The unit string to clean.

    Returns:
        str:
            The cleaned and standardized unit string
    """
    
    # Remove plural 's' and convert to lowercase for unit consistency
    return unit_str.rstrip('s').lower()


def convert_if_whole_number(value):
    """
    Convert a float to an integer if the float is a whole number.

    Args:
        value (float): The float value to be converted.

    Returns:
        int or float: The converted integer if `value` is a whole number, otherwise the original float.

    Example:
        >>> convert_if_whole_number(3.0)
        3
        >>> convert_if_whole_number(3.5)
        3.5
    """
    if value.is_integer():
        return int(value)
    else:
        return value





class ByteConverter:
    """
    A class to convert different units of digital information.

    ...

    Attributes
    ----------
    units : dict
        a dictionary holding the conversion units
    __bytes : int
        a private attribute to hold the value in bytes

    Methods
    -------
    convert(unit)
        Converts the bytes value to the specified unit.
    """
    # Dictionary holding the conversion units
    units = {
        'byte': 1,
        'kilobyte': 1024,
        'megabyte': 1024**2,
        'gigabyte': 1024**3,
        'terabyte': 1024**4,
        'petabyte': 1024**5,
        'exabyte': 1024**6,
        'zetabyte': 1024**7,
        'yottabyte': 1024**8,
        'bit': 1/8,
        'kilobit': 1024/8,
        'megabit': (1024**2)/8,
        'gigabit': (1024**3)/8,
        'terabit': (1024**4)/8,
        'petabit': (1024**5)/8,
        'exabit': (1024**6)/8,
        'zetabit': (1024**7)/8,
        'yottabit': (1024**8)/8,
    }

    def __init__(self, value, unit):
        """
        Constructs all the necessary attributes for the ByteConverter object.

        Parameters
        ----------
            value : int
                the value to be converted
            unit : str
                the unit of the value to be converted
        """
        # Convert the input value to bytes
        self.__bytes = value * self.units[unit]

    def convert(self, unit):
        """
        Converts the bytes value to the specified unit.

        Parameters
        ----------
            unit : str
                the unit to which the value is to be converted

        Returns
        -------
            float
                the converted value
        """
        # Convert the bytes value to the specified unit
        return self.__bytes / self.units[unit]

    # Below are property methods that return the value in a specific unit

    @property
    def bits(self):
        """Returns the value in bits"""
        return self.convert('bit')

    @property
    def kilobits(self):
        """Returns the value in kilobits"""
        return self.convert('kilobit')

    @property
    def megabits(self):
        """Returns the value in megabits"""
        return self.convert('megabit')

    @property
    def gigabits(self):
        """Returns the value in gigabits"""
        return self.convert('gigabit')

    @property
    def terabits(self):
        """Returns the value in terabits"""
        return self.convert('terabit')

    @property
    def petabits(self):
        """Returns the value in petabits"""
        return self.convert('petabit')

    @property
    def exabits(self):
        """Returns the value in exabits"""
        return self.convert('exabit')

    @property
    def zetabits(self):
        """Returns the value in zetabits"""
        return self.convert('zetabit')

    @property
    def yottabits(self):
        """Returns the value in yottabits"""
        return self.convert('yottabit')

    @property
    def bytes(self):
        """Returns the value in bytes"""
        return self.convert('byte')

    @property
    def kilobytes(self):
        """Returns the value in kilobytes"""
        return self.convert('kilobyte')

    @property
    def megabytes(self):
        """Returns the value in megabytes"""
        return self.convert('megabyte')

    @property
    def gigabytes(self):
        """Returns the value in gigabytes"""
        return self.convert('gigabyte')

    @property
    def terabytes(self):
        """Returns the value in terabytes"""
        return self.convert('terabyte')

    @property
    def petabytes(self):
        """Returns the value in petabytes"""
        return self.convert('petabyte')

    @property
    def exabytes(self):
        """Returns the value in exabytes"""
        return self.convert('exabyte')

    @property
    def zetabytes(self):
        """Returns the value in zetabytes"""
        return self.convert('zetabyte')

    @property
    def yottabytes(self):
        """Returns the value in yottabytes"""
        return self.convert('yottabyte')



"""
File: inspyre_toolbox/live_timer/__init__.py
Created: ??

Description:
    inspyre_toolbox.live_timer gives you access to a class and functions that make running a live
    timer that's not only easily to initialize, but easy to query, reset, and pause your timers.
    This module also contains a class named 'TimerHistory' that keeps a history of all timer actions.
    
    To find out more about usage please see:
    
    GIT_REPO_ROOT/examples/live_timer

"""

from time import time

from inspyre_toolbox.core_helpers.logging import ROOT_ISL_DEVICE, add_isl_child
from inspyre_toolbox.live_timer.errors import TimerNotRunningError, TimerNotStartedError
from inspyre_toolbox.live_timer.history import TimerHistory

LOG_NAME = 'InspyreToolbox.live_timer'

LOG = add_isl_child(LOG_NAME)

#ROOT_ISL_DEVICE.adjust_level('debug')

LOG.debug('Log started!')


def format_seconds_to_hhmmss(seconds):
    """
    The format_seconds_to_hhmmss function accepts a number of seconds and returns a string
    representing the amount of time represented by those seconds in HH:MM:SS format.

    Args:
        seconds (int):
            The number of seconds to be converted. (Required)

    Returns:
        A string with the number of hours, minutes and seconds
    """
    hours = seconds // (60 * 60)
    seconds %= (60 * 60)
    minutes = seconds // 60
    seconds %= 60
    return "%02i:%02i:%02i" % (hours, minutes, seconds)


class Timer(object):
    def __repr__(self):
        statement = "Timer(" \
                    f"Started: {self.started} |"
        if self.started:
            runtime = time() - self.start_time
            runtime = format_seconds_to_hhmmss(runtime)
            statement += f" Started: {self.start_time} - Current Runtime: {runtime}"

    def __init__(self, auto_start=False, history=None):

        self.log_name = f'{LOG_NAME}.Timer'

        # Set up logger
        self.log = add_isl_child(self.log_name)

        # Define some default attribute values

        self.running = False
        self.mark_2 = None
        self.pause_end = None
        self.pause_start = time()
        self.paused = False
        self.start_time = None
        self.started = False
        self.stopped = False
        self.total_pause_time = 0
        self.was_paused = False

        self.log.debug('Set up class attributes.')

        # Start a Timer history object to track times for resets
        self.history = TimerHistory(self.__get_elapsed) if history is None else history

        self.log.debug('Timer class instantiated!')

    def __get_elapsed(self, ts=None, sans_pause: bool = False, seconds=False):
        """
        The __get_elapsed function is a private function that is called by the public functions
        start(), pause(), and stop(). It returns the elapsed time in seconds, or as a string if
        seconds=False. The __get_elapsed function does not take any parameters. If you call this
        private function directly, it will return an error

        Args:
            self:
                Access variables that belongs to the class

            ts (int|float):
                Pass in the current timestamp, (Optional, defaults to time.time())

            sans_pause (bool):
                Determine whether or not to include the time that the timer was
                paused in the elapsed time. (Optional; defaults to True)

            seconds (bool):
                Return the elapsed time as number of seconds. (Optional; defaults
                to False)

        Returns:
            The time elapsed since the start of the timer
        """
        log = add_isl_child(f'{self.log_name}__get_elapsed')

        diff_time = self.start_time if ts is None else ts

        # ver1.2.7
        # If we were running but are now stopped, we will skip marking
        if not self.stopped:
            self.mark_2 = time()

        diff = self.mark_2 - diff_time

        if sans_pause:
            return format_seconds_to_hhmmss(diff)

        tpt = 0
        # print(self.total_pause_time)
        if self.paused:
            tpt += time() - self.pause_start

        tpt += self.total_pause_time
        diff = diff - tpt

        return diff if seconds else format_seconds_to_hhmmss(diff)

    def get_elapsed(self, *args, **kwargs):
        """
        The get_elapsed function returns the elapsed time since the timer was started.
        If the timer is not running, it returns None. If it has been stopped,
        it returns the elapsed time from start to stop.

        Args:
            self: Access the attributes and methods of the class in python
            *args: Pass a non-keyworded, variable-length argument list
            **kwargs: Pass a keyworded, variable-length argument list

        Returns:
            The current elapsed time since the timer was started

        """
        if not self.running and self.stopped or self.running:
            self.history.add("QUERY")
            return self.__get_elapsed(*args, **kwargs)
        else:
            try:
                raise TimerNotStartedError(skip_print=True)
            except TimerNotStartedError as e:
                print(e.message)

    def reset(self):
        """

        Reset the 'self.start_time' attribute to this moment, effectively resetting the timer to '00:00:00'

        """
        self.history.add(action="RESET")

        return Timer(history=self.history)

        # if self.running:
        #     self.stop()
        # self.total_pause_time = 0
        # self.pause_start = None
        # self.pause_end = None
        # self.started = False

    def restart(self):
        """
        The restart function resets the timer to its original state and starts it
        running.

        Returns:
            None
        """
        self.reset()
        if not self.started:
            self.start()

    def start(self):
        """

        Store the time the thread was started and assign the attribute 'self.started' to 'True' to indicate this.

        """
        self.start_time = time()
        self.started = True
        self.history.add()
        self.running = True

    def pause(self):
        """

        Pause the running timer.

        This function will pause the running timer.

        What this really means in this context is that this function will fill the
        'self.pause_start' time with the current time. When you unpause at a later
        time the 'self.unpause' function will reference 'self.pause_start' for
        comparison.

        Note:
            This function also sets the 'self.paused' variable to 'True'.

        """
        if not self.started:
            raise TimerNotStartedError()
        if not self.running:
            raise TimerNotRunningError()
        if self.paused:
            return False
        self.pause_start = time()
        self.paused = True
        self.was_paused = True
        self.history.add("PAUSE")

    def unpause(self):
        """

        Un-Pause the running timer.

        This function will unpause the running timer. What that really means in this
        context is that since it marks the time that one paused the timer previously
        and this will compare the time elapsed between when the pause function
        was called and when this is called and keeps track of it.

        Whenever get_elapsed is called, it looks at the amount of seconds that
        have been added by this function's operation process and removes that
        total from the number of seconds elapsed before passing them to be
        naturalized.

        Returns:
            bool:
                If;
                    * True:
                        The timer was successfully un-paused
                    * False:
                        The timer was no successfully un-paused (most likely due to its
                        state not actually being paused),
        """
        if not self.started:
            raise TimerNotStartedError()
        if not self.running:
            raise TimerNotRunningError()
        if self.paused:
            self.pause_end = time()
            diff = self.pause_end - self.pause_start
            self.total_pause_time += diff
            self.paused = False
            self.history.add("UNPAUSE")
            return True
        else:
            return False

    def stop(self):
        """
        The 'stop' function stops the timer, but leaves it in a paused state.
        The 'resume' function resumes the timer from where it was stopped.

        Args:
            self: Reference the object itself

        Returns:
            The time when the timer is stopped
        """
        if not self.started:
            raise TimerNotStartedError()
        if self.running:
            self.running = False
            self.paused = False
            self.stopped = True
            self.mark_2 = time()
        else:
            raise TimerNotRunningError()



#  Copyright (c) 2021. Taylor-Jayde Blackstone <t.blackstone@inspyre.tech> https://inspyre.tech
import os


def add_to_path(program_path: str):
    """
    The add_to_path function adds the program path to the system's PATH variable.
    This allows for easy access to programs from any terminal window.

    Args:
        program_path (str):
            The full path to the directory you'd like to add to your system's PATH
            environment variable.

    Returns:
        None
    """
    if os.name == "nt":  # Windows systems
        import winreg  # Allows access to the Windows registry
        import ctypes  # Allows interface with low-level C APIs

        with winreg.ConnectRegistry(None, winreg.HKEY_CURRENT_USER) as root:  # Get the current user registry
            with winreg.OpenKey(root, "Environment", 0, winreg.KEY_ALL_ACCESS) as key:  # Go to the environment key
                existing_path_value = winreg.EnumValue(key, 3)[1]  # Grab the current path value

                # | v |  Takes the current path value and appends the new program
                # | v |  path
                new_path_value = existing_path_value + program_path + ";"

                # | v |  Update path variable with update path value | v |
                winreg.SetValueEx(key, "PATH", 0, winreg.REG_EXPAND_SZ,
                                  new_path_value)

            # Tell other processes to update their environment
            HWND_BROADCAST = 0xFFFF
            WM_SETTINGCHANGE = 0x1A
            SMTO_ABORTIFHUNG = 0x0002
            result = ctypes.c_long()
            SendMessageTimeoutW = ctypes.windll.user32.SendMessageTimeoutW
            SendMessageTimeoutW(HWND_BROADCAST, WM_SETTINGCHANGE, 0, u"Environment", SMTO_ABORTIFHUNG, 5000,
                                ctypes.byref(result), )
    else:  # If system is *nix
        with open(f"{os.getenv('HOME')}/.bashrc", "a") as bash_file:  # Open bashrc file
            bash_file.write(f'\nexport PATH="{program_path}:$PATH"\n')  # Add program path to Path variable
        os.system(f". {os.getenv('HOME')}/.bashrc")  # Update bash source
    print(f"Added {program_path} to path, please restart shell for changes to take effect")



from os import makedirs
from pathlib import Path
from time import time


class TimerHistory(object):

    def __init__(self, elapsed_method):
        self.get_elapsed = elapsed_method
        self.ledger = []
        self.actions = [
                "START",
                "STOP",
                "PAUSE",
                "UNPAUSE",
                "RESET",
                "CREATE",
                "QUERY"
        ]
        self.add("CREATE")

    def add(self, action: str = "START") -> dict:
        """
        The add function adds a new entry to the ledger.

        Args:
            self:
                Refer to the object instance
            action (str):
                The action you'd like to log to the ledger. (Optional; defaults to 'START')

        Returns:
            The entry that was just added to the ledger

        """
        action = action.upper()
        entry = {
                "time":               time(),
                "elapsed_since_last": "",
                "action":             action,
                "rt_at_create":       ""
        }
        if action == "CREATE":
            entry['elapsed_since_last'] = 0.00
        else:
            entry['elapsed_since_last'] = self.get_elapsed(self.ledger[-1]['time'])
            entry['rt_at_create'] = self.get_elapsed(self.ledger[0]['time'])

        self.ledger.append(entry)

        return entry

    def write(self):
        """
        The write function writes the ledger to a file.

        Args:
            self:
                Refer to the object itself

        Returns:
            None
        """
        data_path = Path("~/Inspyre-Softworks/Inspyre-Toolbox/data").expanduser()

        filename = f'ledger_{str(time()).split(".")[0]}'
        filepath = str(f'{str(data_path)}/{filename}.txt')

        filepath = str(Path(filepath).resolve())

        if not data_path.exists():
            makedirs(data_path)

        with open(filepath, "w") as fp:
            fp.write(str(self.ledger))

    def reset(self) -> None:
        """
        Clear the ledger

        This literally just runs '.clear()' on 'self.ledger'

        Returns:
            None

        """
        self.ledger.clear()



class InvalidPreReleaseTypeError(Exception):

    message = "The pre-release type found in the VERSION file is not valid."

    def __init__(self, message=None, pr_type=None, skip_print=False):
        """
        The __init__ function is the constructor for a class. It is called whenever
        an instance of a class is created.

        The __init__ function can take arguments, but those arguments are only
        stored as attributes on the instance.

        Arguments:
            message (:class:`str`):
                Provide a custom message for the exception. (Optional, defaults to
                ``None``)

            pr_type (:class:`str`):
                The pre-release type found in the VERSION file. (Optional, defaults
                to ``None``)

            skip_print (:class:`bool`)
                Prevent the printing of an error message when the exception is
                raised in a try/except block. (Optional, defaults to ``False``).
        """
        if pr_type:
            self.message = f'The provided pre-release type "{pr_type}" is not valid.'
        if message:
            self.message += f"Addditional information from the caller: {message}"

        if not skip_print:
            print(self.message)




class VersionInfoMismatchError(Exception):
    message = 'The version number does not match the concatenation of the ' \
              'minor, major, and patch numbers.'

    def __init__(
            self,
            version_number: str,
            concatenated: str,
            message=None,
            skip_print=False,
            *args, **kwargs):
        """
        The :class:`VersionInfoMismatchError` class is used to raise an error if the
        version number does not match the concatenation of the minor, major,
        and patch numbers.

        Arguments:
            version_number:
                The version number of the current version of
                                       inspyre_toolbox

        Args:
            self:
            version_number:
            concatenated:
            message:
            skip_print:
            *args:
            **kwargs:

        Returns:

        """
        mismatch_statement = f"Ver. Number: {version_number} | " \
                             f"Concatenated: {concatenated}"
        from_caller = f"Additional information from caller: {message}" if message else ""
        self.message = \
            message or f"{self.message}\n{mismatch_statement}\n{from_caller}"
        super().__init__(self.message, *args, **kwargs)



#  Copyright (c) 2021. Taylor-Jayde Blackstone <t.blackstone@inspyre.tech> https://inspyre.tech
#  Added in v1.1

from inspyre_toolbox.humanize.errors.numerical import NumericalErrors


class HumanizeErrors(object):
    NumericalErrors = NumericalErrors



#  Copyright (c) 2021. Taylor-Jayde Blackstone <t.blackstone@inspyre.tech> https://inspyre.tech
"""

A package containing decorators that are a quick help in programming with Python

"""
import sys
from contextlib import contextmanager
from os import devnull

SUPPRESSED = False


@contextmanager
def suppress_stdout():
    """
    The suppress_stdout function is a context manager that redirects stdout to
    /dev/null.

    This is useful for suppressing output from functions and methods,
    especially when called in loops.  For example:

        with suppress_stdout():
            for i in range(10):
                print(i) # this will not print anything

    Args:
        ``None``
    """
    global SUPPRESSED

    if SUPPRESSED:
        with open(devnull, "w") as dn:
            old_stdout = sys.stdout
            sys.stdout = dn
            try:
                yield
            finally:
                sys.stdout = old_stdout
    else:
        yield



__all__ = [
    'AUTHORS',
    'RELEASE_MAP',
    'URLS',
    'VERSION'
]


URLS = dict(
    developer_url='https://inspyre.tech',
    docs_url='https://inspyre-toolbox.readthedocs.io/en/latest',
    github_url='https://github.com/tayjaybabee/Inspyre-Toolbox',
    pypi_url='https://pypi.org/project/inspyre-toolbox',
)
"""The URLs used in the project."""


AUTHORS = [
    ('Inspyre-Softworks', URLS['developer_url']),
    ('Taylor-Jayde Blackstone', '<t.blackstone@inspyre.tech>')
]
"""The authors of the project."""


RELEASE_MAP = {
    'dev': 'Development Build',
    'alpha': 'Alpha Build',
    'beta': 'Beta Build',
    'rc': 'Release Candidate Build',
    'final': 'Final Release Build'
}
"""The release map for the project."""


VERSION = {
    'major': 1,
    'minor': 4,
    'patch': 0,
    'release': 'final',
    'release_num': 0
}
"""The version information for the project."""



"""
This module provides a command-line interface for converting between different information storage units.
It includes a custom ArgumentParser class that parses the user input, allowing for conversion operations
between various storage units, with 'bytes' as the default unit to convert from.

Usage:
    python ist_bytes_converter.py <amount> [<from_unit>] <to_unit>

Example:
    python ist_bytes_converter.py 1024 bytes KB
    This will convert 1024 bytes to kilobytes.
"""

from argparse import ArgumentParser

# Constants for program name and description
PROG = 'ist-bytes-converter'
DESCRIPTION = 'Convert between information storage units.'

# Exported names
__all__ = ['Arguments']


class Arguments(ArgumentParser):
    """
    Custom ArgumentParser class for parsing command-line arguments for the byte converter program.

    This class inherits from argparse.ArgumentParser and customizes initialization to define specific
    command-line arguments needed for the conversion between information storage units.

    Attributes:
        __parsed (argparse.Namespace, optional): A namespace object containing the parsed arguments.
            It's initially None and gets populated after parsing command-line arguments.
    
    Methods:
        __init__: Initializes the custom ArgumentParser with predefined program name and description,
            and sets up the command-line arguments.
        parsed: Parses the command-line arguments if not already done and returns the parsed namespace.
    """

    def __init__(self, *args, **kwargs):
        """
        Initializes the custom ArgumentParser with the program name, description, and command-line arguments.

        The method sets up the argument parser with a specific program name and description. It defines
        command-line arguments for specifying the amount of storage units to convert, the unit to convert
        from (optional, defaults to 'bytes'), and the unit to convert to.

        Args:
            *args: Variable length argument list for ArgumentParser.
            **kwargs: Arbitrary keyword arguments for ArgumentParser.
        """
        super().__init__(prog=PROG, description=DESCRIPTION, *args, **kwargs)
        
        # Define the command-line arguments for the program
        self.add_argument('amount', type=float, help='The amount of storage units to convert.')
        self.add_argument('from_unit', type=str, nargs='?', default='bytes',
                          help='The storage unit to convert from. Optional, defaults to "bytes".')
        self.add_argument('to_unit', type=str, help='The storage unit to convert to.')
        
        self.__parsed = None  # Placeholder for parsed arguments
    
    @property
    def parsed(self):
        """
        Parses the command-line arguments if not already done and returns the parsed namespace.

        This method checks if the command-line arguments have already been parsed to avoid
        redundant parsing. If not already parsed, it parses the arguments and stores the result
        in the `__parsed` attribute.

        Returns:
            argparse.Namespace: The namespace object containing the parsed arguments.
        """
        if self.__parsed is None:
            self.__parsed = self.parse_args()
        return self.__parsed



#  Copyright (c) 2021. Taylor-Jayde Blackstone <t.blackstone@inspyre.tech> https://inspyre.tech
#  Added in v1.1


class NumericalErrors(object):
    class NotANumberError(Exception):
        """

        Raised when a parameter is provided that is not an integer or a float where one was expectecd.

        """

        def __init__(self, var, var_name: str = None, msg: str = None, skip_print: bool = False, redirect_print=None):
            """

            Raised when a parameter is provided that is not an integer or a float where one was expectecd.

            Arguments:
                var (``Any``): e variable which triggered the raising of this exception.
                var_name (str, optional): The name of the variable that was passed to raise this exception. Defaults to
                None.
                msg (str, optional): The message you'd like the end-user to see upon raising. Defaults to None.
                skip_print (bool, optional): Should we skip printing the message on initialization. Defaults to False.
                redirect_print (any output, optional): An object to redirect our printing to. Ignored if 'skip_print' is
                True. Defaults to None.


            Returns:
                None.

            """
            if var_name is None:
                msg1 = "The variable provided is neither an integer or a float. One or the other is needed."
            else:
                var_type = type(var)
                msg1 = f"{var_name} has a value of {var} and is of type {var_type}. An integer or float was needed."

            msg = msg1 if msg is None else f'{msg1} {msg}'
            self.msg = msg
            self.message = self.msg

            if not skip_print:
                if redirect_print:
                    redirect_print(self.message)
                else:
                    print(self.message)



#  Copyright (c) 2021. Taylor-Jayde Blackstone <t.blackstone@inspyre.tech> https://inspyre.tech
"""

This module contains exceptions for 'InspyreToolbox.proc_man'

"""


class NoFoundProcessesError(Exception):
    """

    Raised when 'InspyreToolbox.proc_man.kill_all_by_name' finds no processes that contain the provided string.

    """

    def __init__(self, *args):
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        if self.message:
            return 'NoFoundProcessesError, {0} '.format(self.message)
        else:
            return 'NoFoundProcessesError has been raised!'



"""

This module contains different functions that help with debugging.

"""
import inspect


def who_rang() -> str:
    """
    The :func:`who_rang` function returns the name of the function that called the
    function that called it.

    Arguments:
        ``None``

    Returns:
        str: The name of the function that called the function that called it.
    """
    return inspect.stack()[2][3]



from inspyre_toolbox.common.meta import VERSION as _VERSION, URLS as _URLS, RELEASE_MAP
import requests
from packaging import version as pkg_version
import sys
from rich.table import Table
from rich import print


def get_full_version_name():
    """
    Gets the full version name.

    Returns:
        str: The full version name.

    Since:
        v1.3.2
    """
    ver = parse_version()
    ver = ver.split('-')[0]

    release_type = RELEASE_MAP[_VERSION["release"]]
    release_num = _VERSION["release_num"]
    release_str = f" {release_type} {'' if _VERSION['release'].lower() == 'final' else f'({release_num})'}"
    return f'v{ver}{release_str}'


def parse_version() -> str:
    """
    Parses the version information into a string.

    Returns:
        str: The version information.

    Since:
        v1.3.2
    """
    version = f'{_VERSION["major"]}.{_VERSION["minor"]}.{_VERSION["patch"]}'

    if _VERSION['release'] != 'final':
        version += f'-{_VERSION["release"]}.{_VERSION["release_num"]}'

    return version


class PyPiVersionInfo:
    """
    A class to represent the version information for this package from PyPi.

    Attributes:
        latest_stable (packaging.version.Version):
            The latest stable version of the package on PyPi.

        latest_pre_release (packaging.version.Version):
            The latest pre-release version of the package.
    Since:
        v1.3.1
    """
    __all_versions: (list[str], None)
    __url = f'{_URLS["pypi_url"]}/json'
    __installed = parse_version()
    __checked_for_update = False
    __newer_available_version = None

    def __init__(self, include_pre_release_for_update_check=False):
        self.__latest_stable = None
        self.__latest_pre_release = None
        self.__all_versions = None
        self.include_pre_release_for_update_check = include_pre_release_for_update_check
        self.__query_versions()

    @property
    def all_versions(self):
        if self.__all_versions is None:
            self.__query_versions()
        return sorted([pkg_version.parse(v) for v in self.__all_versions])

    @property
    def checked_for_update(self):
        return self.__checked_for_update

    @property
    def installed(self):
        """
        Gets the installed version of the package.
        """
        return pkg_version.parse(self.__installed)

    @property
    def installed_newer_than_latest(self):
        return self.installed > self.latest

    @property
    def latest(self):
        return (
            self.all_versions[-1]
            if self.include_pre_release_for_update_check
            else self.latest_stable
        )

    @property
    def latest_stable(self):
        """
        Gets the latest stable version of the package on PyPi.
        """
        if self.__latest_stable is None:
            self.__query_versions()
        return pkg_version.parse(self.__latest_stable)

    @property
    def latest_pre_release(self):
        """
        Gets the latest pre-release version of the package.
        """
        if self.__all_versions is None:
            self.__query_versions()
        pre_release_versions = [v for v in self.all_versions if v.is_prerelease]
        return pre_release_versions[-1] if pre_release_versions else None

    @property
    def newer_available_version(self):

        if self.__newer_available_version is None:
            self.check_for_update()

        return self.__newer_available_version

    def __query_versions(self):
        """
        Queries the versions from PyPi.
        """
        try:
            response = requests.get(self.__url)
            response.raise_for_status()
            data = response.json()

            self.__all_versions = list(data['releases'].keys())
            self.__latest_stable = data['info']['version']
        except requests.RequestException as e:
            # Handle connection errors, HTTP errors, etc.
            print(f"Error querying PyPi: {e}")

    @property
    def update_available(self):
        """
        Checks if an update is available.
        """
        if self.__newer_available_version is None:
            self.check_for_update()
        return self.__newer_available_version is not None

    def check_for_update(self, include_pre_releases=False):
        latest_version = self.latest_stable

        if include_pre_releases or self.include_pre_release_for_update_check:
            latest_version = max(latest_version, self.latest_pre_release)

        if latest_version > self.installed:
            self.__newer_available_version = latest_version

        self.__checked_for_update = True

        return latest_version > self.installed

    def update(self):
        """
        Checks for updates to inspy-logger.

        Returns:
            None`

        Since:
            v1.3.2
        """

        local_newer_statement = 'Local version is newer than latest version. This is likely a development build.'

        if not self.installed_newer_than_latest:
            local_newer_statement = ''

        try:
            if self.update_available:
                print(f'\n\n[bold green]Update Available![/bold green] New version: '
                      f'[bold cyan]{self.update_available}[/bold cyan]')
            else:
                print(f'\n\n[bold green]No update available.[/bold green] Current version: '
                      f'[bold cyan]{parse_version()}[/bold cyan] {local_newer_statement}')

        except Exception as e:
            print(f'An error occurred during the update check: {str(e)}')

    def print_version_info(self):
        """
        Print version information in a formatted table.

        This function creates a table using the `Table` class and populates it with version information about the current Python environment. It then prints the table to the console.

        Parameters:
            self: The current instance of the class.

        Returns:
            None
        """

        # Create a table

        table = Table(show_header=False, show_lines=True, expand=True, border_style='bright_blue',
                      row_styles=['none', 'dim'])

        # Add columns
        table.add_column('Property', style='cyan', width=20)
        table.add_column('Value', justify='center')

        # Add rows
        table.add_row('Version', parse_version(), )
        table.add_row('Full Version Name', get_full_version_name())

        if self.update_available:
            table.add_row('Update Available', '[bold green]Yes[/bold green]')
            table.add_row('Latest Version', f'{self.new_version_available_num}')

        table.add_row('Python Executable Path', sys.executable)
        table.add_row('Python Version', sys.version)

        print(table)



"""
Inspyre-Toolbox Generations Module

This module is part of the Inspyre-Toolbox project. It provides functionalities
for generating random decimal numbers within a specified range. The main function
in this module, `generate_random_decimal`, can generate a specified number of
random decimal numbers, each within a given range.

This module can be useful for various applications, such as creating random
datasets for testing, generating noise for simulations, or any other situation
where random decimal numbers are needed.

Function:

generate_random_decimal(number_of_results=1, minimum=0, maximum=1, force_list=False):
Generates a list of random decimal numbers within a specified range.
This module is released under The MIT License.

Project: Inspyre-Toolbox
Author: Inspyre Softworks - https://inspyre.tech
Created: 5/11/2023 @ 7:48 PM
File:
  Name: init.py
  Path: inspyre_toolbox/generations
"""
import random


def generate_random_decimal(number_of_results=1, minimum=0, maximum=1, force_list=False):
    """
    Generates a list of random decimal numbers within a specified range.

    This function generates a list of random decimal numbers, each of which is
    between the specified minimum and maximum values. If only one number is
    requested, the function returns that number directly unless 'force_list'
    is set to True, in which case a list containing that number is returned.

    Arguments:
        number_of_results (int, optional): The number of random numbers to generate.
            (default: 1)
        minimum (float, optional): The lower bound for the random numbers.
            (default: 0)
        maximum (float, optional): The upper bound for the random numbers.
            (default: 1)
        force_list (bool, optional): Whether to always return a list, even when
            only one number is generated. (default: False)

    Returns:
        float or list of float: A list of random decimal numbers. If only one
            number is generated and 'force_list' is False, that number is
            returned directly.
    """
    res = [random.uniform(minimum, maximum) for _ in range(number_of_results)]
    return res[0] if number_of_results == 1 and not force_list else res



"""
The MIT License (MIT)
Copyright © 2023 Inspyre Softworks - https://inspyre.tech
Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE."""
